/*
 * File operations performed by an Ext2-compatible filesystem.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <e2access/access.h>
#include <e2access/image.h>
#include <e2access/path.h>

#include "ext2_file_operations.h"

/* Object type tests. */

bool Ext2FileOperations::is_directory(ext2_ino_t ino_file)
{
    std::lock_guard<std::mutex> guard(_lock);

    return _image_isdir(_fs, ino_file);
}

bool Ext2FileOperations::is_file(ext2_ino_t ino_file)
{
    std::lock_guard<std::mutex> guard(_lock);

    return _image_isfile(_fs, ino_file);
}

/* Create a file in the directory indicated by the given inode number with the
   given filename. The file is created with the given user permissions. */

long Ext2FileOperations::create_file(ext2_ino_t ino_parent, const char *filename,
                                     user_t user, ext2_ino_t *ino)
{
    if (!path_is_leafname(filename))
        return -L4_EINVAL;

    std::lock_guard<std::mutex> guard(_lock);

    struct ext2_inode inode_parent;
    errcode_t retval = ext2fs_read_inode(_fs, ino_parent, &inode_parent);

    if (retval)
        return -L4_EIO;

    if (!access_can_write(user, &inode_parent))
        return -L4_EPERM;

    if (image_create_file(_fs, ino_parent, filename, 0666 & ~user.umask,
                          user.uid, user.gid, ino))
        return -L4_EIO;

    return L4_EOK;
}

/* For the given 'path', return an inode number or indicate the 'remaining'
   part of the path that cannot be resolved. */

long Ext2FileOperations::find_file(const char *path, ext2_ino_t *ino,
                                   const char **remaining)
{
    std::lock_guard<std::mutex> guard(_lock);

    *remaining = path;
    errcode_t retval = image_find_path(_fs, remaining, ino);

    // NOTE: Map error conditions.

    if (retval)
        return -L4_EIO;

    return L4_EOK;
}

/* Perform closing operations on a file. */

void Ext2FileOperations::close_file(ext2_file_t file)
{
    std::lock_guard<std::mutex> guard(_lock);

    ext2fs_file_flush(file);
    ext2fs_file_close(file);
}

/* Open the file associated with the indicated inode. */

long Ext2FileOperations::open_file(ext2_ino_t ino, ext2_file_t *file)
{
    std::lock_guard<std::mutex> guard(_lock);

    errcode_t retval = ext2fs_file_open(_fs, ino, EXT2_FILE_WRITE, file);

    // NOTE: Map error conditions.

    if (retval)
        return -L4_EIO;

    return L4_EOK;
}

/* Obtain the size of a file. */

offset_t Ext2FileOperations::get_size(ext2_file_t file)
{
    std::lock_guard<std::mutex> guard(_lock);

    return ext2fs_file_get_size(file);
}

/* Update the size of a file. */

void Ext2FileOperations::set_size(ext2_file_t file, offset_t size)
{
    std::lock_guard<std::mutex> guard(_lock);

    ext2fs_file_set_size(file, size);
}

/* Populate the given memory region with file content. */

offset_t Ext2FileOperations::read_file(ext2_file_t file, offset_t filepos, void *addr, offset_t size)
{
    std::lock_guard<std::mutex> guard(_lock);

    unsigned int nread;

    ext2fs_file_llseek(file, filepos, SEEK_SET, NULL);
    ext2fs_file_read(file, (void *) addr, size, &nread);

    return (offset_t) nread;
}

/* Transfer content from the given memory region to a file. */

void Ext2FileOperations::write_file(ext2_file_t file, offset_t filepos, const void *addr, offset_t size)
{
    std::lock_guard<std::mutex> guard(_lock);

    ext2fs_file_llseek(file, filepos, SEEK_SET, NULL);
    ext2fs_file_write(file, addr, size, NULL);
}

/* Initiate iteration over a directory, with the given 'func' being called
   with directory entry details for each entry. */

long Ext2FileOperations::directory_iterate(ext2_ino_t dir,
                                           int func(struct ext2_dir_entry *, int, int,
                                                    char *, void *),
                                           void *priv_data)
{
    if (ext2fs_dir_iterate(_fs, dir, 0, 0, func, priv_data))
        return -L4_EIO;

    return L4_EOK;
}

long Ext2FileOperations::read_inode(ext2_ino_t ino_file, struct ext2_inode *inode)
{
    std::lock_guard<std::mutex> guard(_lock);

    if (ext2fs_read_inode(_fs, ino_file, inode))
        return -L4_EIO;

    return L4_EOK;
}

// vim: tabstop=4 expandtab shiftwidth=4
